import glob
import logging
import os
import urllib.parse

import coloredlogs

from elbo.connector import ElboConnector
from elbo.resources import ElboResources
from elbo.utils.misc_utils import remove_prefix
from elbo.utils.net_utils import upload_file, download_file

logger = logging.getLogger("elbo.ai.client")
coloredlogs.install(level="DEBUG", logger=logger, fmt="%(name)s %(message)s")


def cp_file_to_elbo(elbo_connector, filename, source_object, destination_dir):
    if not os.path.exists(filename):
        logger.error(f"{filename} does not exist, please check path.")
        exit(0)
    file_size = os.path.getsize(filename)

    destination_file_path = os.path.join(destination_dir, os.path.basename(filename))

    url_response = elbo_connector.get_upload_url(
        file_size, destination_file_path, is_training_task=False
    )
    if url_response is None:
        logger.error(f"is unable to authenticate with server..")
        exit(-6)

    (
        upload_url,
        user_id,
        authorization_token,
        session_id,
        show_low_balance_alert,
        user_first_name,
    ) = url_response
    if (
            upload_url is not None
            and user_id is not None
            and authorization_token is not None
    ):
        destination_dir = remove_prefix(destination_dir, "/")
        rel_path = os.path.relpath(filename, start=source_object)
        if rel_path == ".":
            bucket_key = os.path.join(
                user_id, destination_dir, os.path.basename(source_object)
            )
            bucket_key_for_display = os.path.join(
                destination_dir, os.path.basename(source_object)
            )
        else:
            bucket_key = os.path.join(
                user_id, destination_dir, os.path.basename(source_object), rel_path
            )
            bucket_key_for_display = os.path.join(
                destination_dir, os.path.basename(source_object), rel_path
            )

        if not bucket_key.startswith(user_id):
            logger.error(
                f"something is wrong with the destination {bucket_key}, "
                f"please email this bug report to hi@elbo.ai"
            )
            exit(0)

        logger.info(f"uploading {filename} -> elbo://{bucket_key_for_display} ...")
        _ = upload_file(
            filename, upload_url, user_id, authorization_token, bucket_key=bucket_key,
            bucket_key_for_display=bucket_key_for_display
        )


def cp_to_elbo(elbo_connector, file_path, destination_dir):
    """
    Copy the local file path to elbo storage
    :param elbo_connector: The elbo connector
    :param file_path: The local file path
    :param destination_dir: The elbo path
    :return:
    """
    if os.path.isabs(file_path):
        source_object = file_path
    else:
        source_object = os.path.join(os.getcwd(), file_path)

    if not os.path.exists(file_path):
        logger.error(f"Unable to find {file_path}, does it exist?")
        return 0

    if destination_dir == "." or destination_dir == "/":
        # User intends to copy to root
        destination_dir = ""

    if os.path.isdir(source_object):
        for filename in glob.iglob(
                os.path.join(source_object, "**/**"), recursive=True
        ):
            if os.path.isdir(filename):
                continue
            cp_file_to_elbo(elbo_connector, filename, source_object, destination_dir)
    else:
        cp_file_to_elbo(elbo_connector, file_path, source_object, destination_dir)


def cp_from_elbo(elbo_connector, file_path, destination_dir):
    """
    Copy the local file path to elbo storage
    :param elbo_connector: The elbo connector
    :param file_path: The remote elbo file path
    :param destination_dir: The local file path
    :return:
    """
    if os.path.exists(destination_dir) and not os.path.isdir(destination_dir):
        logger.error(f"{destination_dir} already exists, skipping ... ")

    os.makedirs(destination_dir, exist_ok=True)

    params = {"file_or_dir_path": file_path}

    response = elbo_connector.request(ElboResources.RESOURCE_ENDPOINT, params=params)
    if response is None:
        logger.info(f"Could not retrieve {file_path}")
        return

    artifact_auth = response["client_artifact_auth"]
    artifact_url = urllib.parse.unquote(response["client_url"])
    local_dir_name = download_file(
        artifact_url, destination_dir, artifact_auth, None
    )
    if local_dir_name is not None and os.path.exists(local_dir_name):
        logger.info(f"downloaded to: {local_dir_name}")


if __name__ == "__main__":
    _elbo_connector = ElboConnector()
    cp_from_elbo(_elbo_connector, "elbo://tox.ini", ".")



import logging
import subprocess
import sys
import uuid

import coloredlogs
import requests

logger = logging.getLogger("elbo.client")
coloredlogs.install(level="DEBUG", logger=logger, fmt="%(name)s %(message)s")


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def say_hello(user_first_name=None):
    """
    Say hello to the user
    :param user_first_name: The user's first name
    :return: None
    """
    # TODO: Add fortune? Random geetings?
    if user_first_name is None:
        logger.info(f"Hey there 👋")
    else:
        logger.info(f"Hey {user_first_name} 👋, welcome!")


def generate_short_rand_string():
    """
    Generate a short random string
    :return: The random string
    """
    return str(uuid.uuid4())[:8]


def get_task_id_from_file_name(file_name):
    """
    Get the task id from the file name
    :param file_name: The file name
    :return: The task id
    """
    try:
        task_id = file_name.split("-")[2].split(".")[0]
    except IndexError as _:
        task_id = None

    return task_id


def is_elbo_outdated():
    output = subprocess.check_output([sys.executable, "-m", "pip", "show", "elbo"])
    output = str(output, encoding="utf-8").split("\n")
    installed_version = 0
    for line in output:
        if "Version:" in line:
            installed_version = line.split(" ")[1]
    try:
        response = requests.get(f"https://pypi.org/pypi/elbo/json")
        latest_version = response.json()["info"]["version"]
    except requests.exceptions.ConnectionError as _:
        return False, installed_version
    return installed_version != latest_version, latest_version


def exit_handler():
    is_outdated, latest_version = is_elbo_outdated()
    if is_outdated:
        logger.warning(f"A new version of elbo is available, please install using:")
        logger.warning(f"pip3 install elbo=={latest_version}")


def log():
    _logger = logging.getLogger("elbo.client")
    coloredlogs.install(level="DEBUG", logger=_logger, fmt="%(name)s %(message)s")
    return _logger



import hashlib
import logging
import os
import time
import warnings
from http import HTTPStatus

import paramiko
import questionary
import requests
from cryptography.utils import CryptographyDeprecationWarning
from halo import halo
with warnings.catch_warnings():
    warnings.filterwarnings('ignore', category=CryptographyDeprecationWarning)
    from paramiko.ssh_exception import (
        BadHostKeyException,
        AuthenticationException,
        SSHException,
    )
from tqdm import tqdm

import coloredlogs
from tqdm.utils import CallbackIOWrapper

from elbo.resources import ElboResources
from elbo.utils.file_utils import extract_file
from elbo.utils.misc_utils import get_task_id_from_file_name

logger = logging.getLogger("elbo.client")
coloredlogs.install(level="DEBUG", logger=logger, fmt="%(name)s %(message)s")


def wait_for_ssh_ready(client, ip, password):
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    spinner = halo.Halo(
        text="elbo.client is waiting for node to be ready ...",
        spinner="bouncingBall",
        placement="left",
    )
    print("")
    spinner.start()
    while True:
        try:
            client.connect(hostname=ip, username="root", port=2222, password=password)
            spinner.stop()
            break
        except BadHostKeyException as _:
            time.sleep(4)
        except AuthenticationException as _:
            time.sleep(2)
        except SSHException as _:
            time.sleep(4)
        except paramiko.ssh_exception.NoValidConnectionsError as _:
            time.sleep(4)


def download_file(download_url, download_directory, authorization=None, file_name=None):
    """
    TODO: This is copied over from common/utils in elbo-server Repo. Need to expose this in a common
    package

    Download the file at URL to the download directory. This stores the file in memory and writes to file system.
    This can be used for really large files which may not fit in memory.

    :param file_name:
    :param authorization: The download authorization token
    :param download_url: The download URL
    :param download_directory: The download directory
    :return: Path to file if download is successful, None otherwise
    """
    if not download_url:
        logger.error(f"No download URL")
        return

    if file_name is None:
        file_name = download_url.split("/")[-1]
        if file_name is None:
            logger.error(f"Could not infer filename from {download_url}")
            return None
    local_filename = os.path.join(download_directory, file_name)
    headers = {}
    if authorization is not None:
        headers = {"Authorization": authorization}

    try:
        with requests.get(download_url, stream=True, headers=headers) as r:
            r.raise_for_status()
            total_size_in_bytes = int(r.headers.get("content-length", 0))
            chunk_size = 8192
            progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
            with open(local_filename, "wb") as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    if chunk:
                        f.write(chunk)
                        progress_bar.update(len(chunk))
            progress_bar.close()
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == HTTPStatus.NOT_FOUND:
            logger.warning(f"{download_url} file does not exists.")
            logger.warning(f"Perhaps the task is not complete yet?")
        else:
            logger.error(f"Unable to request {download_url} exception - {e}")
        return None

    if os.path.exists(local_filename):
        return local_filename
    else:
        return None


# TODO: Enable once we get this working
# @cli.command(name='logs')
# @click.option('--task_id',
#              required=True,
#              hide_input=False)
def logs(elbo_connector, task_id):
    """
    Show logs from the task.
    """
    # TODO:
    logger.info(f"Getting logs of - {task_id}...")
    params = {"task_id": task_id}
    response = elbo_connector.request(ElboResources.LOGS_ENDPOINT, params=params)
    if response is not None:
        logger.info(response)


def get_real_time_logs(server_ip):
    # noinspection HttpUrlsUsage
    log_address = f"http://{server_ip}/stream"
    connection_errors = 0
    while True:
        try:
            request = requests.get(log_address, stream=True)
            if request.encoding is None:
                request.encoding = "utf-8"

            for line in request.iter_lines(decode_unicode=True):
                if line and line != "0":
                    logger.info(f"{server_ip}> {line}")
            time.sleep(2)
        except requests.exceptions.ConnectionError as _:
            # Connection errors can happen when SSH is trying to get established
            print(".", end="")
            time.sleep(5)
            connection_errors = connection_errors + 1
            if connection_errors > 10:
                print("*")
                break
            pass
        except Exception as _e:
            logger.error(f"Hit {_e}")
            break


def download_task_artifacts(
    artifact_url, output_directory, artifact_auth, artifact_file
):
    local_file_name = download_file(
        artifact_url, output_directory, artifact_auth, artifact_file
    )
    if local_file_name is None:
        return None

    extracted_dir = extract_file(local_file_name)

    # Check if WandB directory exists, if it does, try to sync WandB
    workspace_dir = os.path.join(extracted_dir, ElboResources.WORKSPACE_DIRECTORY)
    files = os.listdir(workspace_dir)
    if ElboResources.WANDB_DIRECTORY in files:
        sync_wandb = questionary.confirm(
            "detected weights and biases offline data, "
            "do you want to sync them to wandb.ai?",
            qmark="elbo.client",
        ).ask()
        if sync_wandb:
            try:
                current_dir = os.getcwd()
                os.chdir(extracted_dir)
                fix_symlinks_command = (
                    f'find {extracted_dir}  -lname "*" -exec sh -c \'ln -snf '
                    f'"{extracted_dir}/$(readlink "$1")" "$1"\' sh ' + "\\{\\} \\;"
                )
                os.system(fix_symlinks_command)
                wandb_offline_dir = os.path.join(workspace_dir, "wandb/offline-run-*")
                wandb_command = f"wandb sync --include-offline {wandb_offline_dir}"
                os.system(wandb_command)
                os.chdir(current_dir)
            except Exception as _e:
                logger.error(f"unable to sync wandb -- {_e}")

    return extracted_dir


def upload_post_with_retry(file_path, file_size, upload_url, headers):
    """
    Post the upload with retry and status update
    :param file_path: The file path
    :param file_size: The file size
    :param upload_url: The upload URL
    :param headers: The headers
    :return:
    """
    with open(file_path, "rb") as f:
        with tqdm(total=file_size, unit="B", unit_scale=True, unit_divisor=1024) as t:
            wrapped_file = CallbackIOWrapper(t.update, f, "read")
            count = 0
            while True:
                try:
                    if count == 5:
                        logger.error(f"too many connection errors...")
                        break
                    upload_response = requests.post(
                        upload_url, headers=headers, data=wrapped_file
                    )
                    return upload_response
                except requests.exceptions.ConnectionError as _:
                    # This can happen if we are uploading large files, wait for some time...
                    count = count + 1
                    logger.warning(f"Connection error, retrying ...")
                    time.sleep(10)

    return None


def upload_file(file_path, upload_url, user_id, authorization_token, bucket_key=None,
                bucket_key_for_display=None):
    """
    Upload a file to the URL specified
    :param file_path: The file to upload
    :param upload_url: The upload URL
    :param user_id: The user id
    :param authorization_token: The auth token
    :param bucket_key: The user bucket key
    :param bucket_key_for_display: The bucket key that we should show to user
    :return: The bucket key and the SHA256 hash of the uploaded file
    """
    content_type = "application/tar+gzip"
    file_name = os.path.basename(file_path)
    file_size = os.stat(file_path).st_size
    upload_response = None
    file_hash = None
    if file_size > ElboResources.LARGE_FILE_SIZE:
        # Large file, use part upload
        size_of_part = ElboResources.LARGE_FILE_UPLOAD_CHUNK_SIZE
        total_bytes_sent = 0

        part_no = 1
        pbar = tqdm(total=file_size, unit="B", unit_scale=True)
        while total_bytes_sent < file_size:
            if (
                file_size - total_bytes_sent
            ) < ElboResources.LARGE_FILE_UPLOAD_CHUNK_SIZE:
                size_of_part = file_size - total_bytes_sent

            with open(file_path, "rb") as fd:
                fd.seek(total_bytes_sent)
                file_data = fd.read(size_of_part)

            part_hash = hashlib.sha1(file_data).hexdigest()
            headers = {
                "Authorization": authorization_token,
                "X-Bz-File-Name": bucket_key,
                "Content-Type": content_type,
                "X-Bz-Part-Number": str(part_no),
                "Content-Length": str(size_of_part),
                "X-Bz-Content-Sha1": part_hash,
            }

            upload_response = requests.post(upload_url, headers=headers, data=file_data)
            if upload_response.status_code != 200:
                break
            total_bytes_sent = total_bytes_sent + size_of_part
            part_no += 1
            pbar.update(size_of_part)
        pbar.close()
    else:
        if bucket_key is None:
            bucket_key = os.path.join(user_id, file_name)

        with open(file_path, mode="rb") as fd:  # b is important -> binary
            file_contents = fd.read()

        file_hash = hashlib.sha1(file_contents).hexdigest()

        headers = {
            "Authorization": authorization_token,
            "X-Bz-File-Name": bucket_key,
            "Content-Type": content_type,
            "X-Bz-Content-Sha1": file_hash,
            "X-Bz-Info-Author": "None",
            "X-Bz-Server-Side-Encryption": "AES256",
        }

        upload_response = upload_post_with_retry(
            file_path, file_size, upload_url, headers
        )

    if upload_response is not None and upload_response.status_code == 200:
        task_id = get_task_id_from_file_name(file_name)
    else:
        logger.error(f"failed task upload - {upload_response}")
        return None

    if bucket_key_for_display is not None:
        logger.info(f"uploaded to elbo://{bucket_key_for_display}")
    else:
        logger.info(f"uploaded to elbo")

    return bucket_key, file_hash, task_id



import json
import logging

import coloredlogs
import questionary
from inquirer.themes import GreenPassion, term

logger = logging.getLogger("elbo.client")
coloredlogs.install(level="DEBUG", logger=logger, fmt="%(name)s %(message)s")


class ElboTheme(GreenPassion):
    def __init__(self):
        super().__init__()
        self.Question.mark_color = term.yellow
        self.Question.brackets_color = term.pink
        self.Question.default_color = term.yellow
        self.Checkbox.selection_color = term.bold_black_on_pink
        self.Checkbox.selection_icon = "❯"
        self.Checkbox.selected_icon = "◉"
        self.Checkbox.selected_color = term.pink
        self.Checkbox.unselected_color = term.normal
        self.Checkbox.unselected_icon = "◯"
        self.List.selection_color = term.bold_black_on_pink
        self.List.selection_cursor = "❯"
        self.List.unselected_color = term.normal


def prompt_user(compute_options):
    """
    Prompt the user with compute options and get the selection
    :param compute_options: The list of compute options
    :return: The selected option
    """
    options = []
    mapping = {}

    def get_sorting_key(x):
        x_parsed = json.loads(x[1])
        return x_parsed["cost"]

    compute_options = dict(
        sorted(compute_options.items(), key=lambda x: get_sorting_key(x))
    )
    for k, v in compute_options.items():
        v = json.loads(v)
        if v.get("spot") is True:
            suffix = "(spot)"
        else:
            suffix = ""
        k = k.replace(" (spot)", "")
        provider = v["provider"]

        if "linode" in provider.lower():
            provider = f"{provider} (~ 9 mins to provision) (billed hourly)"
        option = (
            f" ${v['cost']:7.4f}/hour {k:>26} {v['num_cpu']:>3} cpu {v['mem']:>5}Gb mem "
            f"{v['gpu_mem']:>4}Gb gpu-mem {provider} {suffix}"
        )
        options.append(option)
        mapping[option] = v

    if len(options) > 1:
        logger.info(f"number of compute choices - {len(options)}")
        chosen = questionary.select(
            "Please choose:",
            choices=options,
        ).ask()
        if chosen is None:
            logger.warning(f"None selected, exiting.")
            exit(0)
            return None
        chosen_compute_type = mapping[chosen]
    elif len(options) == 0:
        logger.error(
            f"was unable to find compute options at the moment. Please email support@elbo.ai if this continues."
        )
        return None
    else:
        chosen_compute_type = list(mapping.values())[0]

    return chosen_compute_type



import logging
import datetime as dt
import time
from time import strftime

import coloredlogs
from dateutil import parser

logger = logging.getLogger("elbo.ai.client")
coloredlogs.install(level="DEBUG", logger=logger, fmt="%(name)s %(message)s")


def format_date(date):
    if date is None or date == "":
        return ""
    if type(date) is int:
        date = dt.datetime.fromtimestamp(date / 1e3)
        return date.strftime("%m/%d/%Y, %H:%M:%S")
    if len(date) > 0:
        # Convert and print date time in local timezone
        return parser.parse(date).astimezone().strftime("%m/%d/%y %H:%M")

    logger.warning(f"Invalid date found - [{date}]")
    return ""


def format_time(time_delta_in_seconds):
    if time_delta_in_seconds is None:
        return ""

    if type(time_delta_in_seconds) is str:
        if len(time_delta_in_seconds) == 0:
            return ""
        try:
            time_delta_in_seconds = int(time_delta_in_seconds)
        except ValueError as _:
            try:
                from tzlocal import get_localzone  # $ pip install tzlocal

                local_tz = get_localzone()
                date_time = dt.datetime.strptime(
                    time_delta_in_seconds, "%H:%M:%S.%f"
                ).replace(tzinfo=local_tz)
                time_delta_in_seconds = date_time.timestamp()
                time_delta_in_seconds = strftime(
                    "%Hh:%Mm:%Ss", time.localtime(time_delta_in_seconds)
                )
                return time_delta_in_seconds
            except ValueError as _:
                return time_delta_in_seconds

    if type(time_delta_in_seconds) is float:
        time_delta_in_seconds = int(time_delta_in_seconds)

    time_delta_in_seconds = strftime("%Hh:%Mm:%Ss", time.gmtime(time_delta_in_seconds))
    return time_delta_in_seconds


def transform_date():
    return lambda x: format_date(x)



import base64
import json
import logging
import os

import coloredlogs
import warnings
from cryptography.utils import CryptographyDeprecationWarning
with warnings.catch_warnings():
    warnings.filterwarnings('ignore', category=CryptographyDeprecationWarning)
    import paramiko
from halo import halo

from elbo.api import ElboRestApi
from elbo.resources import ElboResources
from elbo.ui import prompt_user
from elbo.utils.file_utils import get_temp_file_path, create_tar_gz_archive, read_config
from elbo.utils.misc_utils import say_hello
from elbo.utils.net_utils import wait_for_ssh_ready, upload_file

logger = logging.getLogger("elbo.ai.client")
coloredlogs.install(level="DEBUG", logger=logger, fmt="%(name)s %(message)s")


def request_task_run(elbo_connector, bucket_key, file_hash, task_config, session_id):
    """
    Request the receiver to run the task.

    :param elbo_connector: The elbo connector
    :param bucket_key: The bucket key - The file path in the Bucket
    :param file_hash: The file hash. The receiver will check if the file hash on the Bucket is the same as specified
    :param task_config: The task config provided by the user in the YAML file
    :param session_id: The session id here.
    :return: None
    """
    params = {
        "session_id": session_id,
        "bucket_key": bucket_key,
        "file_hash": file_hash,
        "task_config": base64.b64encode(bytes(json.dumps(task_config), "utf-8")),
    }

    response = elbo_connector.request(ElboResources.SCHEDULE_ENDPOINT, params=params)
    return response


def run_internal(elbo_connector, config):
    if not os.path.exists(config):
        logger.error(f"is unable to find '{config}', is the path correct?")
        exit(-1)

    task_config = read_config(config)
    if task_config is None:
        exit(-2)

    if "name" in task_config:
        logger.info(f"is starting '{task_config['name']}' submission ...")
    else:
        logger.error(f"please specify a task `name` in {config} file.")
        exit(-3)

    # Archive files
    temp_file_name = get_temp_file_path(tgz=True)
    # Get directory path relative to config file
    config_directory = os.path.dirname(config)
    task_dir = os.path.join(config_directory, task_config["task_dir"])
    create_tar_gz_archive(temp_file_name, task_dir)
    logger.info(f"is uploading sources from {task_dir}...")
    file_size = os.path.getsize(temp_file_name)
    if file_size > ElboResources.LARGE_FILE_SIZE:
        logger.error(f"the code artifacts are too large - {file_size} bytes...")
        logger.error(f"please consider excluding unnecessary files from sources..")
        exit(-4)

    url_response = elbo_connector.get_upload_url(file_size=file_size)
    if url_response is None:
        logger.error(f"is unable to authenticate with server..")
        exit(-6)
    (
        upload_url,
        user_id,
        authorization_token,
        session_id,
        show_low_balance_alert,
        user_first_name,
    ) = url_response
    if user_first_name:
        say_hello(user_first_name)

    if show_low_balance_alert:
        #
        # Allow user to continue scheduling. The job will complete even if balance is low
        #
        logger.warning(
            f"The balance on your account is too low, please deposit funds 🙏"
        )

    ip = None
    password = None
    if (
        upload_url is not None
        and user_id is not None
        and authorization_token is not None
    ):
        file_name = os.path.basename(temp_file_name)
        bucket_key = os.path.join(
            ElboResources.TASK_SUBMISSION_PREFIX, user_id, file_name
        )
        bucket_key, file_hash, task_id = upload_file(
            temp_file_name, upload_url, user_id, authorization_token, bucket_key=bucket_key,
            bucket_key_for_display=bucket_key
        )
        api = ElboRestApi()
        if bucket_key is not None and file_hash is not None:
            spinner = halo.Halo(
                text="elbo.client is finding compute options (this may take a while)",
                spinner="bouncingBall",
                placement="left",
            )
            print("")
            spinner.start()
            response = request_task_run(
                elbo_connector, bucket_key, file_hash, task_config, session_id
            )
            spinner.stop()
            if response is not None:
                chosen_type = prompt_user(response)
                spinner = halo.Halo(
                    text="elbo.client is provisioning compute (usually takes ~ 4 minutes, ☕️ time!)",
                    spinner="bouncingBall",
                    placement="left",
                )
                spinner.start()
                response_json = api.provision_compute(
                    chosen_type, session_id, task_config, config
                )
                spinner.stop()
                if response_json is not None:
                    ip, password = process_compute_provisioned_response(
                        response_json, ssh_only=False
                    )
                    client = paramiko.SSHClient()
                    wait_for_ssh_ready(client, ip, password)
                else:
                    logger.error(
                        f"something went wrong with provisioning, please try again ..."
                    )
            else:
                logger.error(f"is unable to get compute options from ELBO servers")
        else:
            logger.info(
                f":(. something went wrong with upload, please send us a bug report at bugs@elbo.ai"
            )

    else:
        if user_id is None:
            logger.error(f"is unable to verify your membership.")
            logger.error(
                "please obtain your token from https://elbo.ai/welcome, and run `elbo login`"
            )
        else:
            logger.info(
                f"is unable to obtain upload url. Please report this to bugs@elbo.ai."
            )

    return ip, password


# noinspection HttpUrlsUsage
def process_compute_provisioned_response(response_json, ssh_only=False):
    """
    Process the response from the server for provisioned compute
    :param ssh_only: Is this task SSH only?
    :param response_json: The response json
    :return: None
    """
    response = response_json
    server_ip = response["ip"]
    task_id = response["task_id"]
    password = response["password"]

    logger.info(f"compute node ip {server_ip}")
    #
    # TODO: Could we route the traffic through CloudFare so we get https endpoint?
    # One problem may be the use of port 2222 for SSH which will not work if we proxy through CloudFare
    #
    if not ssh_only:
        logger.info(f"task with ID {task_id} is submitted successfully.")
        logger.info("----------------------------------------------")
        logger.info(f"ssh using - ssh root@{response['ip']} -p 2222")
        logger.info(f"scp using - scp root@{response['ip']} -p 2222")
        logger.info(f"password: {password}")
        logger.info("----------------------------------------------")
        print("")
        logger.info(f"here are URLS for task logs ...")
        logger.info(f"setup logs        - http://{server_ip}/setup")
        logger.info(f"requirements logs - http://{server_ip}/requirements")
        logger.info(f"task logs         - http://{server_ip}/task")
        print("")
        logger.info(f"TIP: 💡 see task details with command: `elbo show {task_id}`")
        print("")
        logger.info(f"⏳ It may take a minute or two for the node to be reachable.")
    else:
        logger.info("----------------------------------------------")
        logger.info(f"ssh using - ssh root@{response['ip']} -p 2222")
        logger.info(f"scp using - scp root@{response['ip']} -p 2222")
        logger.info(f"password: {password}")
        logger.info("----------------------------------------------")
        print("")
        logger.info(
            f"TIP: 💡 copy your SSH public key to the server for password less login using:"
        )
        logger.info(f"ssh-copy-id -p 2222 root@{response['ip']}")
        print("")
        logger.info(f"TIP: 💡 cancel task using command: `elbo kill {task_id}`")
        print("")
        logger.info(f"⏳ It may take a minute or two for the node to be reachable.")

    return server_ip, password



import json
import logging
import os
from http import HTTPStatus

import coloredlogs
import requests
from tinynetrc import Netrc

from elbo.resources import ElboResources

# noinspection PyMethodMayBeStatic

logger = logging.getLogger("elbo.ai.client")
coloredlogs.install(level="DEBUG", logger=logger, fmt="%(name)s %(message)s")


class ElboConnector:
    # TODO: Add back SSL in Cloud fare after fixing HTTP error 524 -- Long polling
    # noinspection HttpUrlsUsage
    ELBO_HOST = "https://prod.elbo.ai"
    # noinspection HttpUrlsUsage
    ELBO_STAGING_HOST = "http://carry.elbo.ai"
    ELBO_TEST_HOST = "http://localhost:5007"

    def __init__(self):
        self._host_name = ElboConnector.get_elbo_host()

    @staticmethod
    def get_elbo_host():
        if os.getenv("ELBO_TEST_MODE") is not None:
            return ElboConnector.ELBO_TEST_HOST
        elif os.getenv("ELBO_STAGING_MODE") is not None:
            return ElboConnector.ELBO_STAGING_HOST
        else:
            return ElboConnector.ELBO_HOST

    def get_auth_token(self):
        """
        Get the ELBO service authentication tokens using NETRC
        :return: The auth tokens
        """
        netrc = None
        try:
            netrc = Netrc()
        except FileNotFoundError as _:
            logger.error(
                f"could not find ~/.netrc file where auth tokens are stored. "
                f"Please run `elbo login` again."
            )
            exit(-1)
        host_name = self._host_name
        tokens = netrc.get(host_name)
        if tokens["password"] is None:
            logger.error(f"please login to elbo service by running `elbo login`")
            exit(0)
        else:
            return tokens["password"]

    def request(self, end_point, params=None, host=None, method="GET"):
        """
        Call the rest API get
        """
        if not host:
            host = self._host_name
        url = host + "/" + end_point
        token = self.get_auth_token()
        headers = {
            # Flask does not like `_` in header keys
            "TOKEN": token
        }

        try:
            # Keep a high timeout
            if method == "GET":
                response = requests.get(
                    url, headers=headers, params=params, timeout=10 * 60
                )
            elif method == "POST":
                response = requests.post(
                    url, headers=headers, data=params, timeout=10 * 60
                )
            else:
                logger.error(f"{method} not found for submission of request")
                response = None
                exit(-1)

        except requests.exceptions.ConnectionError as _:
            logger.error(f"unable to connect to ELBO Server, please try again...")
            return None
        except requests.exceptions.ReadTimeout as _:
            logger.error(
                f"timed out while trying to connect to the node. "
                f"Usually the node starts execution, please"
                f" check status using `elbo show`"
            )
            return None

        if response.ok:
            response_json = response.text
            response_code = response.status_code
            if response_code == HTTPStatus.NOT_MODIFIED:
                logger.warning(f"task already cancelled or completed.")
                exit(0)
            response = json.loads(response_json)
        else:
            if response.status_code == HTTPStatus.PAYMENT_REQUIRED:
                logger.error(f"🚨 {response.text}")
            elif response.status_code == HTTPStatus.UNAUTHORIZED:
                logger.error(
                    f"🚨 we are not able to authorize you, could you try `elbo login` again?"
                )
                exit(-1)
            elif response.status_code == HTTPStatus.NOT_FOUND:
                logger.error(f"🚨 got resource not found error")
            elif response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
                print("")
                logger.error(f"🚨 we hit a glitch --> {response.text}")
                logger.error(f"Please email contact@elbo.ai")
            else:
                logger.error(
                    f"received unknown error code - "
                    f"{response.status_code} with message - {response.text}"
                )
            response = None

        return response

    def get_upload_url(
        self, file_size, destination_file_path=None, is_training_task=True
    ):
        """
        Get the upload URL for the token
        """
        if not is_training_task:
            is_file_upload_request = 1
        else:
            is_file_upload_request = 0

        if file_size > ElboResources.LARGE_FILE_SIZE:
            # Anything greater than 1G is large file upload
            is_large_file_upload = 1
        else:
            is_large_file_upload = 0

        params = {
            "file_upload": is_file_upload_request,
            "large_file": is_large_file_upload,
        }

        if destination_file_path is not None:
            params["file_name"] = destination_file_path

        if is_training_task:
            response = self.request(ElboResources.UPLOAD_URL_ENDPOINT, params=params)
        else:
            response = self.request(ElboResources.FILE_RECEIVER, params=params)

        if response is None:
            return None

        upload_url = response.get("uploadUrl")
        user_id = response.get("user_id")
        authorization_token = response.get("authorizationToken")
        session_id = response.get("session_id")
        show_low_balance_alert = response.get("add_low_balance_alert")
        user_first_name = response.get("user_first_name")
        return (
            upload_url,
            user_id,
            authorization_token,
            session_id,
            show_low_balance_alert,
            user_first_name,
        )



import os


class ElboResources:
    #
    # Server endpoints
    #
    UPLOAD_URL_ENDPOINT = "upload_url"
    FILE_RECEIVER = "file_receiver"
    SCHEDULE_ENDPOINT = "schedule"
    PROVISION_ENDPOINT = "provision"
    CREATE_ENDPOINT = "create"
    LOGS_ENDPOINT = "logs"
    RESOURCE_ENDPOINT = "resource"
    DASHBOARD_ENDPOINT = "dashboard"
    SIGNUP_ENDPOINT = "signup"
    STATUS_ENDPOINT = "status"
    TASKS_ENDPOINT = "tasks"
    SHOW_ENDPOINT = "show"
    BALANCE_ENDPOINT = "balance"
    REMOVE_ENDPOINT = "rm"
    CANCEL_ENDPOINT = "cancel"
    EXCEPTION_REPORT_ENDPOINT = "exception_report"
    WANDB_DIRECTORY = "wandb"
    WORKSPACE_DIRECTORY = "workspace"
    OBJECT_INFO = "object_info"

    TASK_SUBMISSION_PREFIX = "submissions"

    # 100Mb
    LARGE_FILE_SIZE = 1024 * 1024 * 100
    # 4 Mb
    LARGE_FILE_UPLOAD_CHUNK_SIZE = 4 * 1024 * 1024

    ELBO_TOKEN_FILE = os.path.expanduser("~/.elbo/token")
    ELBO_CACHE_DIR = os.path.expanduser("~/.elbo/cache")
